local famBags = LibStub("AceAddon-3.0"):GetAddon("famBags")
local lbf = LibStub("LibButtonFacade", true)
famBags.Buttons = {
   Current = {}
}
famBags.Buttons.LastBuild = ''
-- :::::::::::: LOCALS ::::::::::::: -------------------------------------------------------------------------

local function create_item_data(button)
   button.info = {
      texture = nil,
      bag = nil,
      slot = nil,
      count = nil,
      quality = nil,
      locked = nil,
      readable = nil,
   }
end

local function partial_sort_bag(b1, b2)
   local a, b = b1.info, b2.info
   if a.id and not b.id then
      return true
   elseif b.id and not a.id then
      return false
   elseif a.type and b.type and a.type ~= b.type then
      return a.type < b.type
   elseif a.subType and b.subType and a.subType ~= b.subType then
      return a.subType < b.subType
   elseif a.bag ~= b.bag then
      return a.bag < b.bag
   else
      return a.slot < b.slot
   end
end

local function full_sort_bag(b1, b2)
   --if type(b1) ~= 'table' or type(b2) ~= 'table' then return nil end
   local a, b = b1.info, b2.info
   if a.id and not b.id then
      return true
   elseif b.id and not a.id then
      return false
   elseif a.quality ~= b.quality then
      return (a.quality or -1) > (b.quality or -1)
   elseif a.isQuestItem ~= b.isQuestItem then
      return a.isQuestItem
   elseif a.type and b.type and a.type ~= b.type then
      return a.type < b.type
   elseif a.subType and b.subType and a.subType ~= b.subType then
      return a.subType < b.subType
   elseif a.bag ~= b.bag then
      return a.bag < b.bag
   else
      return a.slot < b.slot
   end
end

-- :::::::::::: famBags.Buttons ::::::::::::: -------------------------------------------------------------------------
function famBags.Buttons:Create(bag, slot, force)
   if not self[bag] then
      local parent = tContains(famBags.BagIndexes.bags, bag) and famBags.famBagsFrame or famBags.famBankFrame
      self[bag] = CreateFrame("Frame", string.format("famBagsBag_%s", tostring(bag)), parent)
      self[bag]:SetID(bag)
   end
   local b = self[bag][slot]
   if not b or force then
      local name = string.format("famBagsButton_%s_%s", tostring(bag), tostring(slot))
      local size = famBags.db.profile.layout.buttons.size
      b = b or CreateFrame("CheckButton", name, self[bag], "ContainerFrameItemButtonTemplate")
      b:SetScale(size)

      self[bag][slot] = b
      local cd = _G[name.."Cooldown"]
      cd:SetAllPoints(b)
      cd:SetFrameLevel(10)
      
      b.quality = b.quality or CreateFrame("Frame", name..'_quality', b)
      b.quality.texture = b.quality.texture or b.quality:CreateTexture(nil, 'OVERLAY')
      famBags.Buttons:UpdateQualityGlow(b)
      b.quality:SetAllPoints()

      if famBags.db.profile.layout.buttons.indicators_enabled then
         if bag == KEYRING_CONTAINER then
            if not b.keyIndicator then
               b.keyIndicator = CreateFrame("Frame", name..'_indicator', b)
               local t = b.keyIndicator:CreateTexture(nil, 'OVERLAY')
               t:SetTexture('Interface\\Addons\\famBags\\Textures\\key.blp')
               t:SetBlendMode('ALPHAKEY')
               t:SetAllPoints()
            end
            self:UpdateIndicator(b, b.keyIndicator)
            b.keyIndicator:Show()
         else
            if not b.stackIndicator then
               b.stackIndicator = CreateFrame("Frame", name..'_indicator', b)
               local t = b.stackIndicator:CreateTexture(nil, 'OVERLAY')
               t:SetTexture('Interface\\Addons\\famBags\\Textures\\stack.blp')
               t:SetBlendMode('ALPHAKEY')
               t:SetAllPoints()
            end
            self:UpdateIndicator(b, b.stackIndicator)
            b.stackIndicator:Hide()
         end
      end
      
      b.stacks = b.stacks or {}
      b.info = b.info or {}
      if lbf and famBags.BFBags then
         famBags.BFBags:AddButton(b,{
            Icon = _G[name.."IconTexture"]
         })
      end
   end

   b:SetID(slot)
   if b.stacks then
      wipe(b.stacks)
   end
--   frame:SetScript("PreClick",famBags.Buttons:PreClick)
   if not famBags.CHANGING_PROFILE then
      self:UpdateItem(b, bag, slot)
   end
   self:Update(b)
   self:UpdateButtonCooldown(b)

   return b
end

--[[
function famBags.Buttons:GetFreeIndex()
   for i=1, #self do
      if self[i].inUse then
         return i
      end
   end
   return #self + 1
end]]

function famBags.Buttons:GetButton(bag, slot)
   if not (self[bag] and #self[bag]) then return nil end
   if self[bag][slot] and self[bag][slot].info and self[bag][slot].info.slot == slot then return self[bag][slot] end
   for _, b in ipairs(self[bag]) do
      if b.info.slot == slot then return b end
   end
   return nil
end

function famBags.Buttons:UpdateButtonsCooldown()
   for _, indexes in pairs(famBags.BagIndexes) do
      for _, i in ipairs(indexes) do
         if self[i] then
            for _, b in ipairs(self[i]) do
               self:UpdateButtonCooldown(b)
            end
         end
      end
   end
end

function famBags.Buttons:UpdateButtonCooldown(button)
   if not button then return end
   local cd = _G[button:GetName().."Cooldown"];
   local start, duration, enable = GetContainerItemCooldown(button.info.bag, button.info.slot);
   CooldownFrame_SetTimer(cd, start, duration, enable);
   if ( duration > 0 and enable == 0 ) then
      SetItemButtonTextureVertexColor(button, 0.4, 0.4, 0.4);
   end
   if enable then
      cd:Show()
   else
      cd:Hide()
   end
end

function famBags.Buttons:UpdateButtonsScale()
   local size = famBags.db.profile.layout.buttons.size
   for _, indexes in pairs(famBags.BagIndexes) do
      for _, i in ipairs(indexes) do
         if self[i] then
            for _, b in ipairs(self[i]) do
               b:SetScale(size)
            end
         end
      end
   end
end

function famBags.Buttons:OrderBag(container, ordered)
   table.sort(container, ordered and full_sort_bag or partial_sort_bag)
end

function famBags.Buttons:SearchButtonById(section, id)
   for _, b in ipairs(section) do
      if b.info.id == id then return b end
   end
   return nil
end

function famBags.Buttons:BuildCurrentFromBags(bag, ordered, onebag)
   self:WipeCurrent(bag, onebag or (self.LastBuild ~= 'bags'))
   local container
   if onebag then
      self.Current[bag] = self.Current[bag] or {}
      container = self.Current[bag]
   else
      famBags.Sections:BuildBagsSections(bag)
   end
   for _, i in ipairs(famBags.BagIndexes[bag]) do
      if not self[i] then
         famBags:UpdateBagButtons(i)
      end
      local slots = GetContainerNumSlots(i)
      if slots and slots > 0 then
         if not onebag then
            self.Current[bag][i] = self.Current[bag][i] or {}
            container = self.Current[bag][i]
         end
         for j=1, slots do
            local b = self[i][j]
            if famBags.isStackMode then
               local b1 = self:SearchButtonById(container, b.info.id)
               if b1 then
                  table.insert(b1.stacks, b)
               else
                  table.insert(container, b)
               end
            else
               table.insert(container, b)
            end
         end
         if not onebag then
            if (ordered or famBags.isStackMode) then
               self:OrderBag(container, ordered)
            end
            self:UpdateCounts(container)
         end
      end
   end
   if onebag then
      if (ordered or famBags.isStackMode) then
         self:OrderBag(container, ordered)
      end
      self:UpdateCounts(container)
   end
   self.LastBuild = onebag and 'onebag' or 'bags'
end

function famBags.Buttons:GetButtonCategoryBag(bag, info)
   local cat = famBags.Classification:GetButtonCategory(info)
   if cat then 
      --TODO: or {} ??????
      self.Current[bag][cat] = self.Current[bag][cat] or {}
      return self.Current[bag][cat]
   end
   return nil
end

function famBags.Buttons:BuildCurrentFromCategories(bag, ordered)
   self:WipeCurrent(bag, self.LastBuild ~= 'categories')
   famBags.Sections:BuildCategoriesSections(bag)
   local b, container, b1
   for _, i in ipairs(famBags.BagIndexes[bag]) do
      if not self[i] then
         famBags:UpdateBagButtons(i)
      end
      local slots = GetContainerNumSlots(i)
      if slots then
         for j=1, slots do
            b = self[i][j]
            container = self:GetButtonCategoryBag(bag, b.info)
            if container then
               if famBags.isStackMode then
                  b1 = self:SearchButtonById(container, b.info.id)
                  if b1 then
                     table.insert(b1.stacks, b)
                  else
                     table.insert(container, b)
                  end
               else
                  table.insert(container, b)
               end
            end
         end
      end
   end
   for _, container in pairs(self.Current[bag]) do
      if (ordered or famBags.isStackMode) then
         self:OrderBag(container, ordered)
      end
      self:UpdateCounts(container)
   end
   self.LastBuild = 'categories'
end

function famBags.Buttons:UpdateItemLock(button, bag, slot)
   _, _, button.info.locked = GetContainerItemInfo(bag, slot)
end

function famBags.Buttons:UpdateItem(button, bag, slot)
   local info = button.info
   info.bag = bag
   info.slot = slot
   
   local oldid, oldcount = info.id, info.count
   info.id = GetContainerItemID(bag, slot)

   local texture, count, locked, quality, readable, lootable, link = GetContainerItemInfo(bag, slot)
   
   if oldid == info.id and oldcount == count then return end
   
   info.texture, info.count, info.locked, info.quality, info.readable, info.lootable, info.link = texture, count, locked, quality, readable, lootable, link

   if info.id then
      local name, _, _, _, _, type, subType, stack, equipLoc, _, _ = GetItemInfo(info.id)
      local isQuestItem, questID, isActive = GetContainerItemQuestInfo(bag, slot)
      info.name, info.type, info.subType, info.stack, info.equipLoc, info.isQuestItem, info.questID, info.isActive = 
      name, type, subType, stack, equipLoc, isQuestItem, questID, isActive
      info.spell = (GetItemSpell(info.id))
   else
      info.name,
      info.type,
      info.subType,
      info.stack,
      info.equipLoc,
      info.isQuestItem, 
      info.questID, 
      info.isActive,
      info.spell = nil
   end
end

function famBags.Buttons:Update(button)
   if not button.info then return end
   local info = button.info

   local name = button:GetName()
   local count = _G[name.."Count"]

   SetItemButtonTexture(button, info.texture)

   if info.count and info.count > 1 then
      famBags:SetTextLayout(button, count, famBags.db.profile.layout.counts, tostring(info.count), false, false)
      count:Show()
   else
      famBags:SetTextLayout(button, count, famBags.db.profile.layout.counts, '', false, false)
      count:Hide()
   end

   local questTexture = (info.questID and not info.isActive) and TEXTURE_ITEM_QUEST_BANG or (info.questID or info.isQuestItem) and TEXTURE_ITEM_QUEST_BORDER

   if questTexture then
      button.quality.texture:SetTexture(questTexture)
      button.quality.texture:SetBlendMode('BLEND')
      button.quality.texture:SetVertexColor(1,1,1)
      button.quality.texture:SetAlpha(1)
      button.quality.texture:Show()
   elseif button.info.quality and button.info.quality > 1 and famBags.db.profile.layout.buttons.quality_glow ~= 'none' then
      local r, g, b = GetItemQualityColor(button.info.quality)
      button.quality.texture:SetBlendMode('ADD')
      button.quality.texture:SetVertexColor(r, g, b, famBags.db.profile.layout.buttons.quality_glow_alpha)
      button.quality.texture:Show()
   else
      button.quality.texture:Hide()
   end

end

function famBags.Buttons:UpdateQualityGlow(b, update)
   local db = famBags.db.profile.layout.buttons
   if db.quality_glow ~= 'none' then
      b.quality.texture:SetTexture(string.format('Interface\\Addons\\famBags\\Textures\\Glow_%s.tga', db.quality_glow))
      b.quality.texture:SetBlendMode('ADD')
      b.quality.texture:SetPoint('TOPLEFT', b.quality, 'TOPLEFT', db.quality_glow_inset, -db.quality_glow_inset)
      b.quality.texture:SetPoint('BOTTOMRIGHT', b.quality, 'BOTTOMRIGHT', -db.quality_glow_inset, db.quality_glow_inset)
   else
      b.quality.texture:Hide()
   end
   if update then
      self:Update(b)
   end
end

function famBags.Buttons:UpdateAllButtonsQuality()
   for _, indexes in pairs(famBags.BagIndexes) do
      for _, i in ipairs(indexes) do
         if self[i] then
            for _, b in ipairs(self[i]) do
               self:UpdateQualityGlow(b, true)
            end
         end
      end
   end
end

function famBags.Buttons:UpdateLock(button)
   if not button.info then return end
   SetItemButtonDesaturated(button, button.info.locked)
   if button.info.locked and button.stacks and #button.stacks > 0 then
      local index
      for i, b in ipairs(button.stacks) do
         if not b.info.locked then
            index = i
            break
         end
      end
      if index then
         local swap = button.stacks[index]
         button.stacks, swap.stacks = swap.stacks, button.stacks
         swap.stacks[index] = button
         for i=1, button:GetNumPoints() do
            swap:SetPoint(button:GetPoint(i))
         end
         local buttontext = _G[button:GetName().."Count"]
         local swaptext = _G[swap:GetName().."Count"]
         local t = buttontext:GetText()
         buttontext:SetText(swaptext:GetText())
         swaptext:SetText(t)
         if button.stackIndicator and button.stackIndicator:IsVisible() then
            button.stackIndicator:Hide()
            swap.stackIndicator:Show()
         end
         button:Hide()
         swap:Show()
      end
   end
end

function famBags.Buttons:UpdateAllIndicators()
   local indicator, img
   for _, indexes in pairs(famBags.BagIndexes) do
      for _, i in ipairs(indexes) do
         if self[i] then
            for _, b in ipairs(self[i]) do
               if b.info.bag == KEYRING_CONTAINER then
                  indicator = 'keyIndicator'
                  img = 'key'
               else
                  indicator = 'stackIndicator'
                  img = 'stack'
               end
               if famBags.db.profile.layout.buttons.indicators_enabled then
                  if not b[indicator] then
                     b[indicator] = CreateFrame("Frame", b:GetName()..'_indicator', b)
                     local t = b[indicator]:CreateTexture(nil, 'OVERLAY')
                     t:SetTexture('Interface\\Addons\\famBags\\Textures\\'..img..'.blp')
                     t:SetBlendMode('ALPHAKEY')
                     t:SetAllPoints()
                  end
                  self:UpdateIndicator(b, b[indicator])
                  if indicator == 'keyIndicator' then
                     b[indicator]:Show()
                  end
               elseif b[indicator] then
                  b[indicator]:Hide()
               end
            end
         end
      end
   end
end

function famBags.Buttons:UpdateIndicator(b, indicator)
   local db = famBags.db.profile.layout.buttons
   indicator:SetPoint('TOPRIGHT', b, 'TOPRIGHT', -db.indicators_offset, -db.indicators_offset)
   indicator:SetWidth(16 * db.indicators_size)
   indicator:SetHeight(16 * db.indicators_size)
   indicator:SetFrameLevel(12)
end

function famBags.Buttons:UpdateCounts(section)
   for _, b in ipairs(section) do
      local c = _G[b:GetName().."Count"]
      local count = (b.info.id) and (b.info.count or 1) or 1
      if b.stacks then
         if famBags.db.profile.layout.buttons.indicators_enabled and #b.stacks > 0 and b.info.bag ~= KEYRING_CONTAINER then
            b.stackIndicator:Show()
         end
         for i=1, #b.stacks do
            count = count + (b.stacks[i].info.count or 1)
         end
      else
         count = b.info.count or 0
      end
      if count and count > 1 then
         local db = famBags.db.profile.layout.counts
         c:SetText(string.format('|cff%02x%02x%02x%s|r', db.color[1]*255, db.color[2]*255, db.color[3]*255, tostring(count)));
         c:Show()
      else
         c:SetText(nil)
         c:Hide()
      end
   end
end

function famBags.Buttons:ClearAllPoints(bag)
   for _, i in ipairs(famBags.BagIndexes[bag]) do
      if self[i] then
         for j=1, #self[i] do
            self[i][j]:ClearAllPoints()
            self[i][j]:Hide()
         end
      end
   end
end

function famBags.Buttons:WipeCurrent(bag, full)
   if not self.Current[bag] then
      self.Current[bag] = {}
      return
   end
   for key, b in pairs(self.Current[bag]) do
      if b.info then
         if b.stacks then
            wipe(b.stacks)
            if famBags.db.profile.layout.buttons.indicators_enabled and b.stackIndicator then
               b.stackIndicator:Hide()
            end
         end
      else
         for _, b1 in pairs(b) do
            if b1.stacks then
               wipe(b1.stacks)
               if famBags.db.profile.layout.buttons.indicators_enabled and b1.stackIndicator then
                  b1.stackIndicator:Hide()
               end
            end
         end
         wipe(b)
      end
   end
   if full then
      wipe(self.Current[bag])
   end
end

--local function famBags.Buttons:PreClick(button)
--end

--   UpdateAddOnMemoryUsage()
--   local m1 = GetAddOnMemoryUsage('famBags')
--
--   UpdateAddOnMemoryUsage()
--   print(m1 - GetAddOnMemoryUsage('famBags'))
--   print('-----------------------------')





